;;;
;;; fields.scm --- very simple finite field support
;;;

(define (make-field elements op-sum op-prod)
  (vector 'field elements op-sum op-prod))

(define (is-field-object? f)
  (and (vector? f) (eq? 'field (vector-ref f 0))))
(define (field-elements f)
  (vector-ref f 1))
(define (field-sum-op f)
  (vector-ref f 2))
(define (field-prod-op f)
  (vector-ref f 3))

;; n must be prime, else this will not necessarily be a group
(define (make-finite-field n)
  (make-field (list-tabulate n (lambda (i) i))
              (lambda (a b)
                (modulo (+ a b) n))
              (lambda (a b)
                (modulo (* a b) n))))

(define (field-add-group f)
  (make-group (field-elements f) (field-sum-op f)))

(define (field-additive-identity f)
  (group-identity (field-add-group f)))

(define (field-additive-inverse f e)
  (group-inverse (field-add-group f) e))

(define (field-prod-group f)
  (make-group (delete (field-additive-identity f) (field-elements f))
              (field-prod-op f)))

(define distributive?
  (make-threaded-solver #f 1))

(defsolver distributive?
  (lambda (f)
    (for-all? (field-elements f)
              (lambda (a)
                (for-all? (field-elements f)
                          (lambda (b)
                            (for-all? (field-elements f)
                                      (lambda (c)
                                        (and (equal? ((field-prod-op f) a ((field-sum-op f) b c))
                                                     ((field-sum-op f) ((field-prod-op f) a b) ((field-prod-op f) a c)))
                                             (equal? ((field-prod-op f) ((field-sum-op f) a b) c)
                                                     ((field-sum-op f) ((field-prod-op f) a c) ((field-prod-op f) b c))))))))))))

(define field?
  (make-threaded-solver #f 1))

(defsolver field?
  (lambda (f)
    (and (abelian? (field-add-group f))
         (abelian? (field-prod-group f))
         (distributive? f))))


;;; examples

#|
(field? (make-finite-field 3))
;Value: #t

(field? (make-finite-field 4))
;Value: #f

|#